#ifndef NMEA0183_MESSAGE_BLOCK_H
#define NMEA0183_MESSAGE_BLOCK_H
#include<ace/Vector_T.h>
#include<RACE_Message_Block.h>
#include<string>
#include<stdarg.h> // for writing
#include"Navigational_Types.h"

class NMEA0183_Message_Block;

//##ModelId=4291BA1F0280
//##Documentation
//## thrown when a NMEA0183 protocol error occurs 
class NMEA0183_Protocol_Error : public std::exception
{
public:
    //##ModelId=4291BA20000B
    //##Documentation
    //## Constructor - initialized with an error message in printf style
    NMEA0183_Protocol_Error(NMEA0183_Message_Block *block,const char *msg,...) __attribute__ ((format(printf,3,4)));
    //##ModelId=4291BA20000F
    //##Documentation
    //## Copy constructor
    NMEA0183_Protocol_Error(const NMEA0183_Protocol_Error& other);
    //##ModelId=4291BA200020
    virtual ~NMEA0183_Protocol_Error() throw();
    //##ModelId=4291BA200022
    //##Documentation 
    //## Returns the error messages 
    virtual const char *what() const throw();
    //##ModelId=4291BA200029
    //##Documentation 
    //## returns a copy of the errornous message block
    NMEA0183_Message_Block *getMessageBlock();
private:
    //##ModelId=42C3B67000A3
    //##Documentation
    //## reference to the errornous NMEA_Message_Block
    NMEA0183_Message_Block *m_block;
    //##ModelId=4291BA200007
    //##Documentation
    //## space for the error message
    char m_errormsg[256];
};


//##ModelId=428354D403C3
//##Documentation
//## Interpretes an NMEA0183 Sentence
//## provides common methods for accessing the content
//## For generating NMEA0183-sentence is the set - method and setchecksum
//##
//## An alternative would be to have for each field an datablock, 
//## but a derivation may have caching binary values in the management
//## data and convert with the set-method bevor transfering it to the
//## device.
class NMEA0183_Message_Block : public RACE_Message_Block
{
public:
    //##ModelId=428354D403CD
   //##Documentation
   //## a field in the NMEA0193 Sentence
   class Field
   {
   public:
        //##ModelId=428354D5022C
       Field();
        //##ModelId=428354D50234
       Field(const Field& other);
        //##ModelId=428354D50236
       //##Documentation
       //## Constructor with offset and
       //## length in the field in the datablock
       Field(off_t off,size_t len);
        //##ModelId=428354D50239
       //##Documentation
       //## offset in the Datablock
       off_t getOffset();
        //##ModelId=428354D5023A
       //##Documentation
       //## length in the field in the datablock
       size_t getLength();
   private:
        //##ModelId=428354D5022A
       //##Documentation
       //## offset in the Datablock
       off_t m_offset;
        //##ModelId=428354D5022B
       //##Documentation
       //## length in the field in the datablock
       size_t m_len;
   };
    //##ModelId=428354D403CE
   typedef ACE_Vector<Field,32> FieldVector;
   
    //##ModelId=428354D50248
   NMEA0183_Message_Block(size_t len);
    //##ModelId=428354D5024A
   NMEA0183_Message_Block(const char *str);
    //##ModelId=428354D5024C
   //##Documentation
   //## Initialize a Message Block from an other 
   //## duplicates the Data and copies the fieldls
   //## Needed for further Interpretation of a more specific subclass
   NMEA0183_Message_Block(NMEA0183_Message_Block *);

    //##ModelId=428354D50252
   //##Documentation
   //## calls a NMEA0183_Semantik after the visitor/semantic-Pattern
   void CallVisitor(RACE_MessageVisitor *visitor);

    //##ModelId=428354D50254
   //##Documentation
   //## Parses the Sentence
   bool parse();

    //##ModelId=428354D50255
   //##Documentation
   //## remember of a field position
   //## This is called by parse,
   //## but can be called by a initial Message_Assembler
   void add(const Field& f);

    //##ModelId=42EF6192033A
   //##Documentation
   //## This is called by the Message_Assembler to indicate that its read a '*'
   void nextFieldIsCheckSumField();

   //##ModelId=428354D5025C
   //##Documentation
   //## returns the type of the NMEA0183_Message
   std::string getType();

   //##ModelId=42B8090D0363
   //##Documentation
   //## returns the first 2 letters of the first field / gettype
   std::string getTalkerID();

   //##ModelId=42B8090D0364
   //##Documentation
   //## returns the sentenceid (letters 2-5 from the first field)
   std::string getSentenceID();


   //##ModelId=428354D5025D
   //##Documentation
   //## gets a field als string
   std::string getStringField(size_t fieldno) throw (NMEA0183_Protocol_Error);

    //##ModelId=42EF61920349
   //##Documentation
   //## reads an angel from a field
   Angle_inDeg getAngleField(size_t fieldno);

   //##ModelId=428354D5025F
   //##Documentation
   //## gets a field as double
   double getDoubleField(size_t fieldno) throw (NMEA0183_Protocol_Error);

   //##ModelId=42B8090D0365
   //##Documentation
   //## gets a field as double
   int getIntField(size_t fieldno) throw (NMEA0183_Protocol_Error);

   //##ModelId=428354D50266
   //##Documentation
   //## gets an boolean value
   //## y: character expected if true
   //## n: character expexted if false
   //## throws an exception if an other character is expected
   bool getBooleanField(char y,char n,size_t fieldno) throw (NMEA0183_Protocol_Error);

   //##ModelId=42B8090D0373
   //##Documentation
   //## checks a constant field and throws an exception if not match
   void checkfield(const char *shall,size_t fieldno) throw (NMEA0183_Protocol_Error);
   //##ModelId=428354D5026A
   //##Documentation
   //## returns number of fields in the sentence
   size_t getNumberOfFields();

    //##ModelId=428354D50270
   //##Documentation
   //## calculates exclusive XOR of all characters between, but not including
   //## the '$' and '*'
   //## does not change the rd_ptr
   std::string calcCheckSum();

    //##ModelId=42EF6192034B
   //##Documentation
   //## gets the Checksum from a Messages, throws an Error if no Checksum
   //## in the Message
   std::string getCheckSum() throw (NMEA0183_Protocol_Error);

    //##ModelId=42EF6192034C
   //##Documentation
   //## sets the checksum (for generating a Message)
   void setCheckSum();

   //##ModelId=42B8090D0384
   //##Documentation
   //## reads date (year,month,day) from a field
   void getDatefromField(struct tm& erg,size_t fnum) throw (NMEA0183_Protocol_Error);
   //##ModelId=42B8090D0393
   //##Documentation
   //## reads the fields hours, minutes and seconds and fraction from a field
   void getTimefromField(struct tm& erg,double& fragerg,size_t fnum) throw (NMEA0183_Protocol_Error);

   //##ModelId=42B8090D03B2
   //##Documentation
   //## reads a field HHMMSS.FRAQ and gets the YYMMDD from the current date
   //## It is used in GLL and GGA
   ACE_Time_Value getTimefromHHMMSS(size_t fnum) throw (NMEA0183_Protocol_Error);
    //##ModelId=42EF61920359
   //##Documentation
   //## initialize for writing
   //## the first $ is written automaticaly
   void set(const char *msg,...) __attribute__ ((format(printf,2,3)));
 protected:
   //##ModelId=42C3B670007F
   //##Documentation
   //## vector with positions and length of the fields
   FieldVector m_fields;
   //##ModelId=428354D50243
   //##Documentation
   //## index of the Field with the checksum
   size_t m_chksum_idx;
   
};

#endif
